#include <stdlib.h>
#include <stdio.h>
#include "basic.h"
#include "fifo.h"

static int verbose = 0 ;
static int n = 100 ;
static int m = 10 ;

#define ADD(x,y) __sync_add_and_fetch(x,y)

static fifo_t *fifo ;
static int sum ;

typedef struct {
  int id ;
} ctx_t ;

void *reader(void *_p) {
  ctx_t *p = _p ;
  for ( ; ; ) {
    boxed_int_t *r = get(fifo) ;
    if (r == NULL) return NULL ;
    if (verbose) fprintf(stderr,"%i GOT %i\n",p->id,r->v) ;
    int i = r->v ; free_boxed_int(r) ;
    (void)ADD(&sum,i) ;
  }
}

void *writer(void *_p) {
  ctx_t *p = _p ;
  for (int k = 0 ; k < n ; k++) {
    if (verbose) fprintf(stderr,"%i PUT %i\n",p->id,k+1) ;
    put(fifo,alloc_boxed_int(k+1)) ;
  }
  return NULL ;
}

int seq(int n) {
  int sum = 0 ;
  for (int k = 1 ; k <= n ; k++) sum += k ;
  return sum ;
}

int run(int fsz) {
  sum = 0 ;
  fifo = alloc_fifo(fsz) ;

  pthread_t readers[m],writers[m] ;
  ctx_t arg[m+m] ;
  for (int k = 0 ; k < m ; k++) {
    arg[k].id = k ;     arg[m+k].id = m+k ; 
    create_thread(&readers[k],reader,&arg[k]) ;
    create_thread(&writers[k],writer,&arg[m+k]) ;
  }

  for (int k = 0 ; k < m ; k++) {
    (void)join_thread(&writers[k]) ;
  }
  for (int k = 0 ; k < m ; k++) {
    put(fifo,NULL) ;
  }
  for (int k = 0 ; k < m ; k++) {
    (void)join_thread(&readers[k]) ;
  }
  free_fifo(fifo) ;
  return sum ;
}


int main(int argc, char **argv) {
  --argc ; ++argv ;
  while (argc > 1) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else exit(2) ;
    } else {
      break ;
    }
    --argc ; ++argv ;
  }

  if (argc > 0) n = atoi(argv[0]) ;
  printf("%i",m*seq(n)) ;
  for (int k = 1 ; k < 4 ; k++) {
    int r = run(k) ;
    printf(" %i",r) ;
  }
  printf("\n") ;
  return 0 ;
}
